from functools import partial
from re import compile, IGNORECASE
from typing import Dict, List, Union

from snowddl.blueprint import (
    TableBlueprint,
    TableColumn,
    PrimaryKeyBlueprint,
    UniqueKeyBlueprint,
    ForeignKeyBlueprint,
    DataType,
    Ident,
    SchemaObjectIdent,
    TableConstraintIdent,
    build_schema_object_ident,
    SearchOptimizationItem,
    ObjectType,
    AggregationPolicyBlueprint,
    AggregationPolicyReference,
    MaskingPolicyBlueprint,
    MaskingPolicyReference,
    ProjectionPolicyBlueprint,
    ProjectionPolicyReference,
    RowAccessPolicyBlueprint,
    RowAccessPolicyReference,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile
from snowddl.parser.schema import database_json_schema, schema_json_schema

col_type_re = compile(r"^(?P<type>[a-z0-9_]+(\((\d+|int|float)(,(\d+))?\))?)" r"(?P<not_null> NOT NULL)?$", IGNORECASE)


# fmt: off
table_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "minItems": 1,
            "items": {
                "anyOf": [
                    {
                        "type": "string"
                    },
                    {
                        "type": "object",
                        "properties": {
                            "type": {
                                "type": "string"
                            },
                            "default": {
                                "type": "string"
                            },
                            "default_sequence": {
                                "type": "string"
                            },
                            "expression": {
                                "type": "string"
                            },
                            "collate": {
                                "type": "string"
                            },
                            "comment": {
                                "type": "string"
                            }
                        },
                        "required": ["type"],
                        "additionalProperties": False
                    }
                ]
            }
        },
        "cluster_by": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "change_tracking": {
            "type": "boolean"
        },
        "search_optimization": {
            "anyOf": [
                {
                    "type": "boolean"
                },
                {
                    "type": "object",
                    "additionalProperties": {
                        "type": "array",
                        "items": {
                            "type": "string",
                            "minItems": 1,
                        }
                    }
                }
            ]
        },
        "comment": {
            "type": "string"
        },
        "primary_key": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "unique_keys": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1,
            }
        },
        "foreign_keys": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "ref_table": {
                        "type": "string"
                    },
                    "ref_columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "additionalProperties": False
            },
            "minItems": 1
        },
        "is_transient": {
            "type": "boolean"
        },
        "retention_time": {
            "type": "integer"
        },
        "aggregation_policy": {
            "type": "object",
            "properties": {
                "policy_name": {
                    "type": "string"
                },
                "columns": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "required": ["policy_name"],
            "additionalProperties": False
        },
        "masking_policies": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "policy_name": {
                        "type": "string"
                    },
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "required": ["policy_name", "columns"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "projection_policies": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "policy_name": {
                        "type": "string"
                    },
                    "column": {
                        "type": "string",
                    }
                },
                "required": ["policy_name", "column"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "row_access_policy": {
            "type": "object",
            "properties": {
                "policy_name": {
                    "type": "string"
                },
                "columns": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                },
            },
            "required": ["policy_name", "columns"],
            "additionalProperties": False
        },
    },
    "additionalProperties": False
}
# fmt: on


class TableParser(AbstractParser):
    def load_blueprints(self):
        combined_params = {}

        for database_name in self.get_database_names():
            database_params = self.parse_single_file(f"{database_name}/params", database_json_schema)
            combined_params[database_name] = {}

            for schema_name in self.get_schema_names_in_database(database_name):
                schema_params = self.parse_single_file(f"{database_name}/{schema_name}/params", schema_json_schema)

                combined_params[database_name][schema_name] = {
                    "is_transient": database_params.get("is_transient", False) or schema_params.get("is_transient", False),
                    "retention_time": schema_params.get("retention_time"),
                }

        self.parse_schema_object_files("table", table_json_schema, partial(self.process_table, combined_params=combined_params))

    def process_table(self, f: ParsedFile, combined_params: dict):
        column_blueprints = []

        for col_name, col in f.params["columns"].items():
            # Short syntax
            if isinstance(col, str):
                col = {"type": col}

            m = col_type_re.match(col["type"])

            if not m:
                raise ValueError(
                    f"Incorrect short syntax for column [{col_name}] in table [{f.database}.{f.schema}.{f.name}]: {col['type']}"
                )

            # Default for sequence
            if col.get("default_sequence"):
                col_default = build_schema_object_ident(self.env_prefix, col["default_sequence"], f.database, f.schema)
            else:
                col_default = col.get("default")

            column_blueprints.append(
                TableColumn(
                    name=Ident(col_name),
                    type=DataType(m.group("type")),
                    not_null=bool(m.group("not_null")),
                    default=col_default,
                    expression=col.get("expression"),
                    collate=col.get("collate").lower() if col.get("collate") else None,
                    comment=col.get("comment"),
                )
            )

        bp = TableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            columns=column_blueprints,
            cluster_by=f.params.get("cluster_by", None),
            is_transient=f.params.get("is_transient", combined_params[f.database][f.schema].get("is_transient", False)),
            retention_time=f.params.get("retention_time", combined_params[f.database][f.schema].get("retention_time", None)),
            change_tracking=f.params.get("change_tracking", False),
            search_optimization=self.get_search_optimization(f.params.get("search_optimization", False)),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)

        # Constraints

        if f.params.get("primary_key"):
            key_bp = PrimaryKeyBlueprint(
                full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in f.params.get("primary_key")],
                comment=None,
            )

            self.config.add_blueprint(key_bp)

        for columns in f.params.get("unique_keys", []):
            key_bp = UniqueKeyBlueprint(
                full_name=TableConstraintIdent(
                    self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in columns]
                ),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in columns],
                comment=None,
            )

            self.config.add_blueprint(key_bp)

        for fk in f.params.get("foreign_keys", []):
            key_bp = ForeignKeyBlueprint(
                full_name=TableConstraintIdent(
                    self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in fk["columns"]]
                ),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in fk["columns"]],
                ref_table_name=build_schema_object_ident(self.env_prefix, fk["ref_table"], f.database, f.schema),
                ref_columns=[Ident(c) for c in fk["ref_columns"]],
                comment=None,
            )

            self.config.add_blueprint(key_bp)

        # Policies

        if f.params.get("aggregation_policy"):
            policy_name = build_schema_object_ident(
                self.env_prefix, f.params["aggregation_policy"]["policy_name"], f.database, f.schema
            )

            ref = AggregationPolicyReference(
                object_type=ObjectType.TABLE,
                object_name=bp.full_name,
                columns=[Ident(c) for c in f.params["aggregation_policy"].get("columns", [])],
            )

            self.config.add_policy_reference(AggregationPolicyBlueprint, policy_name, ref)

        for mp in f.params.get("masking_policies", []):
            policy_name = build_schema_object_ident(self.env_prefix, mp["policy_name"], f.database, f.schema)

            ref = MaskingPolicyReference(
                object_type=ObjectType.TABLE,
                object_name=bp.full_name,
                columns=[Ident(c) for c in mp["columns"]],
            )

            self.config.add_policy_reference(MaskingPolicyBlueprint, policy_name, ref)

        for pp in f.params.get("projection_policies", []):
            policy_name = build_schema_object_ident(self.env_prefix, pp["policy_name"], f.database, f.schema)

            ref = ProjectionPolicyReference(
                object_type=ObjectType.TABLE,
                object_name=bp.full_name,
                column=Ident(pp["column"]),
            )

            self.config.add_policy_reference(ProjectionPolicyBlueprint, policy_name, ref)

        if f.params.get("row_access_policy"):
            policy_name = build_schema_object_ident(
                self.env_prefix, f.params["row_access_policy"]["policy_name"], f.database, f.schema
            )

            ref = RowAccessPolicyReference(
                object_type=ObjectType.TABLE,
                object_name=bp.full_name,
                columns=[Ident(c) for c in f.params["row_access_policy"]["columns"]],
            )

            self.config.add_policy_reference(RowAccessPolicyBlueprint, policy_name, ref)

    def get_search_optimization(self, search_optimization: Union[Dict[str, List[str]], bool]):
        # Legacy search optimization on the whole table
        if isinstance(search_optimization, bool):
            return search_optimization

        items = []

        # Detailed search optimization on specific columns
        for method, targets in search_optimization.items():
            for t in targets:
                t_parts = t.split(":", 2)

                items.append(
                    SearchOptimizationItem(
                        method=str(method).upper(),
                        target=Ident(t_parts[0]),
                    )
                )

        return items



from snowddl.blueprint import ExternalFunctionBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import dtypes_from_arguments


class ExternalFunctionResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.EXTERNAL_FUNCTION

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW EXTERNAL FUNCTIONS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            full_name = f"{r['catalog_name']}.{r['schema_name']}.{r['name']}({dtypes_from_arguments(r['arguments'])})"

            existing_objects[full_name] = {
                "database": r["catalog_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "arguments": r["arguments"],
                "comment": r["description"],
                "is_secure": r["is_secure"] == "Y",
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ExternalFunctionBlueprint)

    def create_object(self, bp: ExternalFunctionBlueprint):
        query = self._build_create_function(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl(
            "COMMENT ON FUNCTION {full_name:i} IS {comment}",
            {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: ExternalFunctionBlueprint, row: dict):
        query = self._build_create_function(bp)

        if not query.compare_short_hash(row["comment"]):
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl(
                "COMMENT ON FUNCTION {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": query.add_short_hash(bp.comment),
                },
            )

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "DROP FUNCTION {database:i}.{schema:i}.{name:i}({dtypes:r})",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
                "dtypes": dtypes_from_arguments(row["arguments"]),
            },
        )

        return ResolveResult.DROP

    def _build_create_function(self, bp: ExternalFunctionBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE OR REPLACE")

        if bp.is_secure:
            query.append("SECURE")

        query.append(
            "EXTERNAL FUNCTION {full_name:in} (",
            {
                "full_name": bp.full_name,
            },
        )

        for idx, arg in enumerate(bp.arguments):
            query.append_nl(
                "    {comma:r}{arg_name:i} {arg_type:r}",
                {
                    "comma": "  " if idx == 0 else ", ",
                    "arg_name": arg.name,
                    "arg_type": arg.type,
                },
            )

        query.append_nl(")")

        query.append_nl(
            "RETURNS {ret_type:r}",
            {
                "ret_type": bp.returns,
            },
        )

        query.append_nl(
            "API_INTEGRATION = {api_integration:i}",
            {
                "api_integration": bp.api_integration,
            },
        )

        if bp.is_strict:
            query.append_nl("STRICT")

        if bp.is_immutable:
            query.append_nl("IMMUTABLE")

        if bp.headers:
            query.append_nl("HEADERS = (")

            for idx, (header_name, header_value) in enumerate(bp.headers.items()):
                query.append_nl(
                    "    {comma:r}{header_name} = {header_value}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "header_name": header_name,
                        "header_value": header_value,
                    },
                )

            query.append_nl(")")

        if bp.context_headers:
            query.append_nl(
                "CONTEXT_HEADERS = ({context_headers:r})",
                {
                    "context_headers": bp.context_headers,
                },
            )

        if bp.max_batch_rows:
            query.append_nl(
                "MAX_BATCH_ROWS = {max_batch_rows:d}",
                {
                    "max_batch_rows": bp.max_batch_rows,
                },
            )

        if bp.compression:
            query.append_nl(
                "COMPRESSION = {compression}",
                {
                    "compression": bp.compression,
                },
            )

        if bp.request_translator:
            query.append_nl(
                "REQUEST_TRANSLATOR = {request_translator:i}",
                {
                    "request_translator": bp.request_translator,
                },
            )

        if bp.response_translator:
            query.append_nl(
                "RESPONSE_TRANSLATOR = {response_translator:i}",
                {
                    "response_translator": bp.response_translator,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        query.append_nl(
            "AS {url}",
            {
                "url": bp.url,
            },
        )

        return query



from json import loads

from snowddl.blueprint import ObjectType
from snowddl.converter.abc_converter import ConvertResult
from snowddl.converter.abc_schema_object_converter import AbstractSchemaObjectConverter
from snowddl.converter._yaml import YamlLiteralStr
from snowddl.parser.task import task_json_schema


class TaskConverter(AbstractSchemaObjectConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.TASK

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW TASKS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "when": r["condition"],
                "body": r["definition"],
                "schedule": r["schedule"],
                "after": r["predecessors"],
                "warehouse": r["warehouse"],
                "allow_overlapping_execution": r["allow_overlapping_execution"],
                "error_integration": r["error_integration"],
                "config": r["config"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def dump_object(self, row):
        """
        The following parameters are currently not available for conversion due to lack of data
        in SHOW TASKS and DESC TASK commands output:
        - session_params
        - user_task_timeout_ms
        - suspend_task_after_num_failures
        - finalize
        """
        data = {
            "body": YamlLiteralStr(row["body"]),
            "schedule": row["schedule"],
            "after": self._get_after(row),
            "when": row["when"],
            "warehouse": row["warehouse"],
            "user_task_managed_initial_warehouse_size": self._get_user_task_managed_initial_warehouse_size(row),
            "allow_overlapping_execution": self._get_allow_overlapping_execution(row),
            "error_integration": self._get_error_integration(row),
            "config": row["config"],
            "comment": row["comment"],
        }

        object_path = (
            self.base_path / self._normalise_name_with_prefix(row["database"]) / self._normalise_name(row["schema"]) / "task"
        )

        if data:
            self._dump_file(object_path / f"{self._normalise_name(row['name'])}.yaml", data, task_json_schema)
            return ConvertResult.DUMP

        return ConvertResult.EMPTY

    def _get_user_task_managed_initial_warehouse_size(self, row):
        if row["warehouse"]:
            return None

        return "XSMALL"

    def _get_allow_overlapping_execution(self, row):
        if row["allow_overlapping_execution"] == "true":
            return True

        if row["allow_overlapping_execution"] == "false":
            return False

        return None

    def _get_error_integration(self, row):
        if row["error_integration"] == "null":
            return None

        return row["error_integration"]

    def _get_after(self, row):
        if "after" not in row:
            return None

        after = [self._normalise_name_with_prefix(n) for n in loads(row["after"])]

        if len(after) == 0:
            return None

        return after



from snowddl.blueprint import (
    UserBlueprint,
    AccountObjectIdent,
    AuthenticationPolicyBlueprint,
    AuthenticationPolicyReference,
    NetworkPolicyBlueprint,
    NetworkPolicyReference,
    ObjectType,
    SchemaObjectIdent,
    build_role_ident,
    build_default_namespace_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile
from snowddl.parser.business_role import business_role_json_schema


# fmt: off
user_json_schema = {
    "type": "object",
    "additionalProperties": {
        "properties": {
            "login_name": {
                "type": "string"
            },
            "display_name": {
                "type": "string"
            },
            "first_name": {
                "type": "string"
            },
            "last_name": {
                "type": "string"
            },
            "email": {
                "type": "string"
            },
            "disabled": {
                "type": "boolean"
            },
            "password": {
                "type": "string"
            },
            "rsa_public_key": {
                "type": "string"
            },
            "rsa_public_key_2": {
                "type": "string"
            },
            "type": {
                "type": "string",
            },
            "default_warehouse": {
                "type": "string"
            },
            "default_namespace": {
                "type": "string"
            },
            "session_params": {
                "type": "object",
                "additionalProperties": {
                    "type": ["boolean", "number", "string"]
                }
            },
            "business_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "comment": {
                "type": "string"
            },
            "authentication_policy": {
                "type": "string",
            },
            "network_policy": {
                "type": "string",
            },
        },
        "additionalProperties": False
    }
}
# fmt: on


class UserParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file("user", user_json_schema, self.process_user)

    def process_user(self, f: ParsedFile):
        default_warehouse_map = self.get_default_warehouse_map()

        for user_name, user in f.params.items():
            business_roles = []
            default_warehouse = user.get("default_warehouse")

            for business_role_name in user.get("business_roles", []):
                business_roles.append(build_role_ident(self.env_prefix, business_role_name, self.config.BUSINESS_ROLE_SUFFIX))

                if default_warehouse is None:
                    default_warehouse = default_warehouse_map.get(business_role_name)

            full_user_name = AccountObjectIdent(self.env_prefix, user_name)

            if user.get("login_name"):
                # Login name is string, not identifier (!), special characters like '@#!' are permitted
                # But it still requires env_prefix due to unique constraint
                login_name = f"{self.config.env_prefix}{user.get('login_name')}".upper()
            else:
                login_name = str(full_user_name)

            if user.get("display_name"):
                display_name = str(user.get("display_name")).upper()
            else:
                display_name = str(full_user_name)

            bp = UserBlueprint(
                full_name=full_user_name,
                login_name=login_name,
                display_name=display_name,
                first_name=user.get("first_name"),
                last_name=user.get("last_name"),
                email=user.get("email"),
                disabled=user.get("disabled", False),
                password=user.get("password"),
                rsa_public_key=user.get("rsa_public_key").replace(" ", "") if user.get("rsa_public_key") else None,
                rsa_public_key_2=user.get("rsa_public_key_2").replace(" ", "") if user.get("rsa_public_key_2") else None,
                type=str(user.get("type")).upper() if user.get("type") else None,
                default_warehouse=AccountObjectIdent(self.env_prefix, default_warehouse) if default_warehouse else None,
                default_namespace=build_default_namespace_ident(self.env_prefix, user.get("default_namespace"))
                if user.get("default_namespace")
                else None,
                session_params=self.normalise_params_dict(user.get("session_params", {})),
                business_roles=business_roles,
                comment=user.get("comment"),
            )

            self.validate_user_type_properties(bp)
            self.config.add_blueprint(bp)

            # Authentication policy
            if user.get("authentication_policy"):
                policy_name_parts = user.get("authentication_policy").split(".")

                if len(policy_name_parts) != 3:
                    raise ValueError(
                        f"Authentication policy [{user.get('authentication_policy')}] should use fully-qualified identifier <database>.<schema>.<name> for user [{full_user_name}]"
                    )

                policy_name = SchemaObjectIdent(self.env_prefix, *policy_name_parts)

                ref = AuthenticationPolicyReference(
                    object_type=ObjectType.USER,
                    object_name=full_user_name,
                )

                self.config.add_policy_reference(AuthenticationPolicyBlueprint, policy_name, ref)

            # Network policy
            if user.get("network_policy"):
                policy_name = AccountObjectIdent(self.env_prefix, user.get("network_policy"))

                ref = NetworkPolicyReference(
                    object_type=ObjectType.USER,
                    object_name=full_user_name,
                )

                self.config.add_policy_reference(NetworkPolicyBlueprint, policy_name, ref)

            if "NETWORK_POLICY" in bp.session_params:
                raise ValueError(
                    "NETWORK_POLICY in session_params of USER is no longer supported. Please use dedicated [network_policy] parameter instead. Read more: https://docs.snowddl.com/breaking-changes-log/0.33.0-october-2024"
                )

    def get_default_warehouse_map(self):
        default_warehouse_map = {}

        business_role_config = self.parse_single_file("business_role", business_role_json_schema)

        for business_role_name, business_role in business_role_config.items():
            if "warehouse_usage" not in business_role:
                continue

            default_warehouse_map[business_role_name] = business_role["warehouse_usage"][0]

        return default_warehouse_map

    def validate_user_type_properties(self, bp: UserBlueprint):
        if bp.type in ("SERVICE", "LEGACY_SERVICE"):
            if bp.first_name:
                raise ValueError(f"Property [FIRST_NAME] is not allowed for user [{bp.full_name}] with type [{bp.type}]")

            if bp.last_name:
                raise ValueError(f"Property [LAST_NAME] is not allowed for user [{bp.full_name}] with type [{bp.type}]")

        if bp.type == "SERVICE":
            if bp.password:
                raise ValueError(f"Property [PASSWORD] is not allowed for user [{bp.full_name}] with type [{bp.type}]")



from snowddl.blueprint import AuthenticationPolicyBlueprint, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
authentication_policy_json_schema = {
    "type": "object",
    "properties": {
        "authentication_methods": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "mfa_authentication_methods": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "mfa_enrollment": {
            "type": "string"
        },
        "client_types": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "security_integrations": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}
# fmt: on


class AuthenticationPolicyParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files(
            "authentication_policy", authentication_policy_json_schema, self.process_authentication_policy
        )

    def process_authentication_policy(self, f: ParsedFile):
        # As of Oct 2024, no easy way around hardcoding defaults
        # Cannot distinguish missing value and explicitly set default value
        # https://docs.snowflake.com/en/sql-reference/sql/create-authentication-policy
        bp = AuthenticationPolicyBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            authentication_methods=self.normalise_params_list(f.params.get("authentication_methods"))
            if f.params.get("authentication_methods")
            else ["ALL"],
            mfa_authentication_methods=self.normalise_params_list(f.params.get("mfa_authentication_methods"))
            if f.params.get("mfa_authentication_methods")
            else ["PASSWORD", "SAML"],
            mfa_enrollment=f.params.get("mfa_enrollment").upper() if f.params.get("mfa_enrollment") else "OPTIONAL",
            client_types=self.normalise_params_list(f.params.get("client_types")) if f.params.get("client_types") else ["ALL"],
            security_integrations=self.normalise_params_list(f.params.get("security_integrations"))
            if f.params.get("security_integrations")
            else ["ALL"],
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from enum import Enum
from typing import List

from .object_type import ObjectType
from ..model import BaseModelWithConfig


class PermissionModelRuleset(Enum):
    DATABASE_OWNER = {
        "create_managed_access_schema": False,
        "create_database_owner_role": True,
        "create_database_write_role": True,
        "create_database_read_role": True,
        "create_schema_owner_role": False,
        "create_schema_write_role": True,
        "create_schema_read_role": True,
    }

    SCHEMA_OWNER = {
        "create_managed_access_schema": True,
        "create_database_owner_role": False,
        "create_database_write_role": False,
        "create_database_read_role": False,
        "create_schema_owner_role": True,
        "create_schema_write_role": True,
        "create_schema_read_role": True,
    }

    @property
    def create_managed_access_schema(self):
        return self.value["create_managed_access_schema"]

    @property
    def create_database_owner_role(self):
        return self.value["create_database_owner_role"]

    @property
    def create_database_write_role(self):
        return self.value["create_database_write_role"]

    @property
    def create_database_read_role(self):
        return self.value["create_database_read_role"]

    @property
    def create_schema_owner_role(self):
        return self.value["create_schema_owner_role"]

    @property
    def create_schema_write_role(self):
        return self.value["create_schema_write_role"]

    @property
    def create_schema_read_role(self):
        return self.value["create_schema_read_role"]


class PermissionModelCreateGrant(BaseModelWithConfig):
    on: ObjectType


class PermissionModelFutureGrant(BaseModelWithConfig):
    privilege: str
    on: ObjectType


class PermissionModel(BaseModelWithConfig):
    ruleset: PermissionModelRuleset = PermissionModelRuleset.SCHEMA_OWNER
    owner_create_grants: List[PermissionModelCreateGrant] = []
    owner_future_grants: List[PermissionModelFutureGrant] = []
    write_future_grants: List[PermissionModelFutureGrant] = []
    read_future_grants: List[PermissionModelFutureGrant] = []



from snowddl.blueprint import (
    DatabaseIdent,
    SchemaRoleBlueprint,
    SchemaBlueprint,
    SchemaObjectIdent,
    Grant,
    FutureGrant,
    build_role_ident,
)
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class SchemaRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.SCHEMA_ROLE_SUFFIX

    def get_blueprints(self):
        blueprints = []

        for schema_bp in self.config.get_blueprints_by_type(SchemaBlueprint).values():
            schema_permission_model = self.config.get_permission_model(schema_bp.permission_model)

            if schema_permission_model.ruleset.create_schema_owner_role:
                blueprints.append(self.get_blueprint_owner_role(schema_bp))

            if schema_permission_model.ruleset.create_schema_write_role:
                blueprints.append(self.get_blueprint_write_role(schema_bp))

            if schema_permission_model.ruleset.create_schema_read_role:
                blueprints.append(self.get_blueprint_read_role(schema_bp))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_owner_role(self, schema_bp: SchemaBlueprint):
        grants = []
        account_grants = []
        future_grants = []

        schema_permission_model = self.config.get_permission_model(schema_bp.permission_model)

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.DATABASE,
                name=DatabaseIdent(schema_bp.full_name.env_prefix, schema_bp.full_name.database),
            )
        )

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.SCHEMA,
                name=schema_bp.full_name,
            )
        )

        for model_create_grant in schema_permission_model.owner_create_grants:
            grants.append(
                Grant(
                    privilege=f"CREATE {model_create_grant.on.singular}",
                    on=ObjectType.SCHEMA,
                    name=schema_bp.full_name,
                )
            )

        for model_future_grant in schema_permission_model.owner_future_grants:
            future_grants.append(
                FutureGrant(
                    privilege=model_future_grant.privilege,
                    on_future=model_future_grant.on,
                    in_parent=ObjectType.SCHEMA,
                    name=schema_bp.full_name,
                )
            )

        depends_on = set()

        for additional_grant in schema_bp.owner_additional_grants:
            grants.append(additional_grant)

            # Dependency on another schema role
            if additional_grant.on == ObjectType.ROLE and str(additional_grant.name).endswith(self.get_role_suffix()):
                depends_on.add(additional_grant.name)

        for additional_account_grant in schema_bp.owner_additional_account_grants:
            account_grants.append(additional_account_grant)

        bp = SchemaRoleBlueprint(
            full_name=build_role_ident(
                self.config.env_prefix,
                schema_bp.full_name.database,
                schema_bp.full_name.schema,
                self.config.OWNER_ROLE_TYPE,
                self.get_role_suffix(),
            ),
            grants=grants,
            account_grants=account_grants,
            future_grants=future_grants,
            depends_on=depends_on,
        )

        return bp

    def get_blueprint_read_role(self, schema_bp: SchemaBlueprint):
        grants = []
        future_grants = []

        schema_permission_model = self.config.get_permission_model(schema_bp.permission_model)

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.DATABASE,
                name=DatabaseIdent(schema_bp.full_name.env_prefix, schema_bp.full_name.database),
            )
        )

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.SCHEMA,
                name=schema_bp.full_name,
            )
        )

        for model_future_grant in schema_permission_model.read_future_grants:
            future_grants.append(
                FutureGrant(
                    privilege=model_future_grant.privilege,
                    on_future=model_future_grant.on,
                    in_parent=ObjectType.SCHEMA,
                    name=schema_bp.full_name,
                )
            )

        bp = SchemaRoleBlueprint(
            full_name=build_role_ident(
                self.config.env_prefix,
                schema_bp.full_name.database,
                schema_bp.full_name.schema,
                self.config.READ_ROLE_TYPE,
                self.get_role_suffix(),
            ),
            grants=grants,
            future_grants=future_grants,
        )

        return bp

    def get_blueprint_write_role(self, schema_bp: SchemaBlueprint):
        grants = []
        future_grants = []

        schema_permission_model = self.config.get_permission_model(schema_bp.permission_model)

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.DATABASE,
                name=DatabaseIdent(schema_bp.full_name.env_prefix, schema_bp.full_name.database),
            )
        )

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.SCHEMA,
                name=schema_bp.full_name,
            )
        )

        for model_future_grant in schema_permission_model.write_future_grants:
            future_grants.append(
                FutureGrant(
                    privilege=model_future_grant.privilege,
                    on_future=model_future_grant.on,
                    in_parent=ObjectType.SCHEMA,
                    name=schema_bp.full_name,
                )
            )

        bp = SchemaRoleBlueprint(
            full_name=build_role_ident(
                self.config.env_prefix,
                schema_bp.full_name.database,
                schema_bp.full_name.schema,
                self.config.WRITE_ROLE_TYPE,
                self.get_role_suffix(),
            ),
            grants=grants,
            future_grants=future_grants,
        )

        return bp

    def grant_to_future_grant(self, grant: Grant):
        if not grant.on.is_future_grant_supported:
            return None

        if isinstance(grant.name, SchemaObjectIdent):
            return FutureGrant(
                privilege=grant.privilege,
                on_future=grant.on,
                in_parent=ObjectType.SCHEMA,
                name=grant.name.schema_full_name,
            )

        return None



from snowddl.blueprint import ForeignKeyBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class ForeignKeyResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.FOREIGN_KEY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}
        constraints_by_name = {}

        cur = self.engine.execute_meta(
            "SHOW IMPORTED KEYS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            # Constraint for Hybrid tables are handled separately
            if r["comment"] == ObjectType.HYBRID_TABLE.name:
                continue

            if r["fk_name"] not in constraints_by_name:
                constraints_by_name[r["fk_name"]] = {
                    "database": r["fk_database_name"],
                    "schema": r["fk_schema_name"],
                    "table": r["fk_table_name"],
                    "columns": {r["key_sequence"]: r["fk_column_name"]},
                    "ref_database": r["pk_database_name"],
                    "ref_schema": r["pk_schema_name"],
                    "ref_table": r["pk_table_name"],
                    "ref_columns": {r["key_sequence"]: r["pk_column_name"]},
                }
            else:
                constraints_by_name[r["fk_name"]]["columns"][r["key_sequence"]] = r["fk_column_name"]
                constraints_by_name[r["fk_name"]]["ref_columns"][r["key_sequence"]] = r["pk_column_name"]

        for c in constraints_by_name.values():
            columns_list = [c["columns"][k] for k in sorted(c["columns"])]
            ref_columns_list = [c["ref_columns"][k] for k in sorted(c["ref_columns"])]

            full_name = f"{c['database']}.{c['schema']}.{c['table']}({','.join(columns_list)})"

            existing_objects[full_name] = {
                "database": c["database"],
                "schema": c["schema"],
                "table": c["table"],
                "columns": columns_list,
                "ref_table_name": f"{c['ref_database']}.{c['ref_schema']}.{c['ref_table']}",
                "ref_columns": ref_columns_list,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ForeignKeyBlueprint)

    def create_object(self, bp: ForeignKeyBlueprint):
        self.engine.execute_safe_ddl(
            "ALTER TABLE {table_name:i} ADD FOREIGN KEY ({columns:i}) REFERENCES {ref_table_name:i} ({ref_columns:i})",
            {
                "table_name": bp.table_name,
                "columns": bp.columns,
                "ref_table_name": bp.ref_table_name,
                "ref_columns": bp.ref_columns,
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: ForeignKeyBlueprint, row: dict):
        if str(bp.ref_table_name) == row["ref_table_name"] and [str(c) for c in bp.ref_columns] == row["ref_columns"]:
            return ResolveResult.NOCHANGE

        self.engine.execute_safe_ddl(
            "ALTER TABLE {table_name:i} DROP FOREIGN KEY ({columns:i})",
            {
                "table_name": bp.table_name,
                "columns": bp.columns,
            },
        )

        self.engine.execute_safe_ddl(
            "ALTER TABLE {table_name:i} ADD FOREIGN KEY ({columns:i}) REFERENCES {ref_table_name:i} ({ref_columns:i})",
            {
                "table_name": bp.table_name,
                "columns": bp.columns,
                "ref_table_name": bp.ref_table_name,
                "ref_columns": bp.ref_columns,
            },
        )

        return ResolveResult.ALTER

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "ALTER TABLE {database:i}.{schema:i}.{table:i} DROP FOREIGN KEY ({columns:i})",
            {
                "database": row["database"],
                "schema": row["schema"],
                "table": row["table"],
                "columns": row["columns"],
            },
        )

        return ResolveResult.DROP



from snowddl.blueprint import AlertBlueprint, AccountObjectIdent, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
alert_json_schema = {
    "type": "object",
    "properties": {
        "warehouse": {
            "type": "string"
        },
        "schedule": {
            "type": "string"
        },
        "condition": {
            "type": "string"
        },
        "action": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        },
    },
    "required": ["warehouse", "schedule", "condition", "action"],
    "additionalProperties": False,
}
# fmt: on


class AlertParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("alert", alert_json_schema, self.process_alert)

    def process_alert(self, f: ParsedFile):
        bp = AlertBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            warehouse=AccountObjectIdent(self.env_prefix, f.params["warehouse"]),
            schedule=str(f.params["schedule"]).strip(),
            condition=str(f.params["condition"]).strip(),
            action=str(f.params["action"]).strip(),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import Grant, AccountGrant, TechnicalRoleBlueprint, ObjectType, build_role_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
technical_role_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "account_grants": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "additionalProperties": False
    }
}
# fmt: on


class TechnicalRoleParser(AbstractParser):
    def load_blueprints(self):
        # Current config name
        self.parse_single_file("technical_role", technical_role_json_schema, self.process_technical_role)

        # Backwards compatible name
        self.parse_single_file("tech_role", technical_role_json_schema, self.process_technical_role)

    def process_technical_role(self, f: ParsedFile):
        for technical_role_name, technical_role in f.params.items():
            technical_role_ident = build_role_ident(self.env_prefix, technical_role_name, self.config.TECHNICAL_ROLE_SUFFIX)

            grants = []
            account_grants = []

            for definition, pattern_list in technical_role.get("grants", {}).items():
                on, privileges = definition.upper().split(":")

                for p in privileges.split(","):
                    for pattern in pattern_list:
                        blueprints = self.config.get_blueprints_by_type_and_pattern(ObjectType[on].blueprint_cls, pattern)

                        if not blueprints:
                            raise ValueError(f"No {ObjectType[on].plural} matched wildcard grant with pattern [{pattern}]")

                        for object_bp in blueprints.values():
                            grants.append(
                                Grant(
                                    privilege=p,
                                    on=ObjectType[on],
                                    name=object_bp.full_name,
                                )
                            )

            for privilege in technical_role.get("account_grants", []):
                account_grants.append(AccountGrant(privilege=privilege))

            bp = TechnicalRoleBlueprint(
                full_name=technical_role_ident,
                grants=grants,
                account_grants=account_grants,
                comment=technical_role.get("comment"),
            )

            self.config.add_blueprint(bp)



from snowddl.blueprint import (
    FunctionBlueprint,
    AccountObjectIdent,
    Ident,
    SchemaObjectIdentWithArgs,
    ArgumentWithType,
    NameWithType,
    DataType,
    BaseDataType,
    StageWithPath,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
function_json_schema = {
    "type": "object",
    "properties": {
        "language": {
            "type": "string"
        },
        "arguments": {
            "type": "object",
            "minItems": 1,
            "items": {
                "anyOf": [
                    {
                        "type": "string"
                    },
                    {
                        "type": "object",
                        "properties": {
                            "type": {
                                "type": "string"
                            },
                            "default": {
                                "type": "string"
                            },
                        },
                        "required": ["type"],
                        "additionalProperties": False
                    }
                ]
            }
        },
        "returns": {
            "anyOf": [
                {
                    "type": "string"
                },
                {
                    "type": "object",
                    "additionalProperties": {
                        "type": "string"
                    }
                }
            ]
        },
        "body": {
            "type": "string"
        },
        "is_secure": {
            "type": "boolean"
        },
        "is_strict": {
            "type": "boolean"
        },
        "is_immutable": {
            "type": "boolean"
        },
        "is_memoizable": {
            "type": "boolean"
        },
        "runtime_version": {
            "type": ["number", "string"]
        },
        "imports": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "stage": {
                        "type": "string"
                    },
                    "path": {
                        "type": "string"
                    }
                },
                "required": ["stage"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "packages": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "handler": {
            "type": "string"
        },
        "external_access_integrations": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "secrets": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            },
        },
        "comment": {
            "type": "string"
        }
    },
    "anyOf": [
        {
            "required": ["body", "returns"]
        },
        {
            "required": ["handler", "returns"]
        }
    ],
    "additionalProperties": False
}
# fmt: on


class FunctionParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("function", function_json_schema, self.process_function)

    def process_function(self, f: ParsedFile):
        arguments = self.get_arguments(f)
        base_name = self.validate_name_with_args(f.path, arguments)

        bp = FunctionBlueprint(
            full_name=SchemaObjectIdentWithArgs(
                self.env_prefix, f.database, f.schema, base_name, data_types=[a.type.base_type for a in arguments]
            ),
            language=f.params.get("language", "SQL"),
            body=f.params.get("body"),
            arguments=arguments,
            returns=self.get_returns(f),
            is_secure=f.params.get("is_secure", False),
            is_strict=f.params.get("is_strict", False),
            is_immutable=f.params.get("is_immutable", False),
            is_memoizable=f.params.get("is_memoizable", False),
            runtime_version=str(f.params.get("runtime_version")) if f.params.get("runtime_version") else None,
            imports=self.get_imports(f),
            packages=f.params.get("packages"),
            handler=f.params.get("handler"),
            external_access_integrations=self.get_external_access_integrations(f),
            secrets=self.get_secrets(f),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)

    def get_arguments(self, f: ParsedFile):
        arguments = []

        for arg_name, arg in f.params.get("arguments", {}).items():
            # Short syntax
            if isinstance(arg, str):
                arg = {"type": arg}

            arg_with_type = ArgumentWithType(
                name=Ident(arg_name),
                type=DataType(arg["type"]),
                default=str(arg["default"]) if "default" in arg else None,
            )

            if arg_with_type.type.base_type == BaseDataType.VECTOR:
                raise NotImplementedError("Argument data type VECTOR is currently not supported for FUNCTION object type")

            arguments.append(arg_with_type)

        return arguments

    def get_returns(self, f: ParsedFile):
        if isinstance(f.params["returns"], dict):
            # Returns table
            return [NameWithType(name=Ident(k), type=DataType(t)) for k, t in f.params["returns"].items()]

        return DataType(f.params["returns"])

    def get_imports(self, f: ParsedFile):
        if f.params.get("imports"):
            return [
                StageWithPath(
                    stage_name=build_schema_object_ident(self.env_prefix, i["stage"], f.database, f.schema),
                    path=self.normalise_stage_path(i["path"]),
                )
                for i in f.params.get("imports")
            ]

        return None

    def get_external_access_integrations(self, f: ParsedFile):
        if f.params.get("external_access_integrations"):
            return [AccountObjectIdent(self.env_prefix, i) for i in f.params.get("external_access_integrations")]

        return None

    def get_secrets(self, f: ParsedFile):
        if f.params.get("secrets"):
            return {
                k: build_schema_object_ident(self.env_prefix, v, f.database, f.schema) for k, v in f.params.get("secrets").items()
            }

        return None



from snowddl.blueprint import AccountParameterBlueprint, Ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
account_params_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": ["boolean", "number", "string"]
    }
}
# fmt: on


class AccountParameterParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file("account_params", account_params_json_schema, self.process_account_params)

    def process_account_params(self, f: ParsedFile):
        for name, value in f.params.items():
            bp = AccountParameterBlueprint(
                full_name=Ident(name),
                value=value,
                comment=None,
            )

            if str(bp.full_name) == "NETWORK_POLICY":
                raise ValueError(
                    "NETWORK_POLICY in account_params.yaml is no longer supported. Please use account_policy.yaml instead. Read more: https://docs.snowddl.com/breaking-changes-log/0.33.0-october-2024"
                )

            self.config.add_blueprint(bp)



from snowddl.blueprint import ResourceMonitorBlueprint, AccountObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
resource_monitor_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "credit_quota": {
                "type": "integer"
            },
            "frequency": {
                "type": "string"
            },
            "triggers": {
                "type": "object",
                "additionalProperties": {
                    "type": "string"
                }
            }
        },
        "required": ["credit_quota", "frequency", "triggers"],
        "additionalProperties": False
    }
}
# fmt: on


class ResourceMonitorParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file("resource_monitor", resource_monitor_json_schema, self.process_resource_monitor)

    def process_resource_monitor(self, f: ParsedFile):
        for name, monitor in f.params.items():
            bp = ResourceMonitorBlueprint(
                full_name=AccountObjectIdent(self.env_prefix, name),
                credit_quota=int(monitor.get("credit_quota")),
                frequency=str(monitor.get("frequency")).upper(),
                triggers={int(k): str(v).upper() for k, v in monitor.get("triggers").items()},
                comment=None,
            )

            self.config.add_blueprint(bp)



from ._parsed_file import ParsedFile
from ._scanner import DirectoryScanner
from .account_params import AccountParameterParser
from .account_policy import AccountPolicyParser
from .aggregation_policy import AggregationPolicyParser
from .alert import AlertParser
from .authentication_policy import AuthenticationPolicyParser
from .business_role import BusinessRoleParser
from .database import DatabaseParser
from .dynamic_table import DynamicTableParser
from .event_table import EventTableParser
from .external_access_integration import ExternalAccessIntegrationParser
from .external_function import ExternalFunctionParser
from .external_table import ExternalTableParser
from .file_format import FileFormatParser
from .function import FunctionParser
from .hybrid_table import HybridTableParser
from .materialized_view import MaterializedViewParser
from .masking_policy import MaskingPolicyParser
from .network_policy import NetworkPolicyParser
from .network_rule import NetworkRuleParser
from .outbound_share import OutboundShareParser
from .permission_model import PermissionModelParser
from .pipe import PipeParser
from .placeholder import PlaceholderParser
from .procedure import ProcedureParser
from .projection_policy import ProjectionPolicyParser
from .resource_monitor import ResourceMonitorParser
from .row_access_policy import RowAccessPolicyParser
from .schema import SchemaParser
from .secret import SecretParser
from .sequence import SequenceParser
from .stage import StageParser
from .stream import StreamParser
from .table import TableParser
from .task import TaskParser
from .technical_role import TechnicalRoleParser
from .user import UserParser
from .view import ViewParser
from .warehouse import WarehouseParser


default_parse_sequence = [
    AccountParameterParser,
    # --
    AggregationPolicyParser,
    AuthenticationPolicyParser,
    MaskingPolicyParser,
    NetworkPolicyParser,
    ProjectionPolicyParser,
    RowAccessPolicyParser,
    ResourceMonitorParser,
    AccountPolicyParser,
    # --
    WarehouseParser,
    DatabaseParser,
    SchemaParser,
    SecretParser,
    NetworkRuleParser,
    ExternalAccessIntegrationParser,
    FileFormatParser,
    StageParser,
    SequenceParser,
    FunctionParser,
    ExternalFunctionParser,
    ProcedureParser,
    TableParser,
    EventTableParser,
    HybridTableParser,
    DynamicTableParser,
    ExternalTableParser,
    StreamParser,
    MaterializedViewParser,
    ViewParser,
    PipeParser,
    TaskParser,
    AlertParser,
    # --
    OutboundShareParser,
    TechnicalRoleParser,
    BusinessRoleParser,
    UserParser,
]


singledb_parse_sequence = [
    AggregationPolicyParser,
    MaskingPolicyParser,
    ProjectionPolicyParser,
    RowAccessPolicyParser,
    # --
    DatabaseParser,
    SchemaParser,
    SecretParser,
    NetworkRuleParser,
    FileFormatParser,
    StageParser,
    SequenceParser,
    FunctionParser,
    ExternalFunctionParser,
    ProcedureParser,
    TableParser,
    EventTableParser,
    HybridTableParser,
    DynamicTableParser,
    ExternalTableParser,
    StreamParser,
    MaterializedViewParser,
    ViewParser,
    PipeParser,
    TaskParser,
    AlertParser,
]



from argparse import ArgumentParser, HelpFormatter
from os import environ, getcwd
from pathlib import Path
from shutil import rmtree

from snowddl.app.base import BaseApp
from snowddl.blueprint import DatabaseIdent, ObjectType
from snowddl.config import SnowDDLConfig
from snowddl.converter import default_converter_sequence
from snowddl.settings import SnowDDLSettings


class ConvertApp(BaseApp):
    def init_arguments_parser(self):
        formatter = lambda prog: HelpFormatter(prog, max_help_position=32)

        parser = ArgumentParser(
            prog="snowddl-convert",
            description="Convert existing objects in Snowflake account to SnowDDL config",
            formatter_class=formatter,
        )

        # Config
        parser.add_argument(
            "-c", help="Path to output config directory (default: current directory)", metavar="CONFIG_PATH", default=getcwd()
        )

        # Auth
        parser.add_argument(
            "-a",
            help="Snowflake account identifier (default: SNOWFLAKE_ACCOUNT env variable)",
            metavar="ACCOUNT",
            default=environ.get("SNOWFLAKE_ACCOUNT"),
        )
        parser.add_argument(
            "-u",
            help="Snowflake user name (default: SNOWFLAKE_USER env variable)",
            metavar="USER",
            default=environ.get("SNOWFLAKE_USER"),
        )
        parser.add_argument(
            "-p",
            help="Snowflake user password (default: SNOWFLAKE_PASSWORD env variable)",
            metavar="PASSWORD",
            default=environ.get("SNOWFLAKE_PASSWORD"),
        )
        parser.add_argument(
            "-k",
            help="Path to private key file (default: SNOWFLAKE_PRIVATE_KEY_PATH env variable)",
            metavar="PRIVATE_KEY",
            default=environ.get("SNOWFLAKE_PRIVATE_KEY_PATH"),
        )

        # Role & warehouse
        parser.add_argument(
            "-r",
            help="Snowflake active role (default: SNOWFLAKE_ROLE env variable)",
            metavar="ROLE",
            default=environ.get("SNOWFLAKE_ROLE"),
        )
        parser.add_argument(
            "-w",
            help="Snowflake active warehouse (default: SNOWFLAKE_WAREHOUSE env variable)",
            metavar="WAREHOUSE",
            default=environ.get("SNOWFLAKE_WAREHOUSE"),
        )

        # Options
        parser.add_argument(
            "--authenticator",
            help="Authenticator: 'snowflake' or 'externalbrowser' (to use any IdP and a web browser) (default: SNOWFLAKE_AUTHENTICATOR env variable or 'snowflake')",
            default=environ.get("SNOWFLAKE_AUTHENTICATOR", "snowflake"),
        )
        parser.add_argument(
            "--passphrase",
            help="Passphrase for private key file (default: SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable)",
            default=environ.get("SNOWFLAKE_PRIVATE_KEY_PASSPHRASE"),
        )
        parser.add_argument(
            "--env-prefix",
            help="Env prefix added to global object names, used to separate environments (e.g. DEV, PROD)",
            default=environ.get("SNOWFLAKE_ENV_PREFIX"),
        )
        parser.add_argument(
            "--env-prefix-separator",
            help="Custom separator for Env prefix (supported values are: '__', '_', '$')",
            choices=["__", "_", "$"],
            default=environ.get("SNOWFLAKE_ENV_PREFIX_SEPARATOR", "__"),
        )
        parser.add_argument(
            "--max-workers", help="Maximum number of workers to resolve objects in parallel", default=None, type=int
        )
        parser.add_argument("--clean", help="Delete existing config files before conversion", default=False, action="store_true")

        # Logging
        parser.add_argument(
            "--log-level", help="Log level (possible values: DEBUG, INFO, WARNING; default: INFO)", default="INFO"
        )

        # Object types
        parser.add_argument(
            "--exclude-object-types", help="Comma-separated list of object types NOT to convert", default=None, metavar=""
        )
        parser.add_argument(
            "--include-object-types",
            help="Comma-separated list of object types TO convert, all other types are excluded",
            default=None,
            metavar="",
        )

        # Target specific database only
        parser.add_argument("--include-databases", help="Comma-separated list of databases to convert", default=None, metavar="")
        parser.add_argument(
            "--ignore-ownership",
            help="Ignore OWNERSHIP of databases and schemas during conversion process, makes it possible to convert objects owned by another role",
            default=False,
            action="store_true",
        )

        parser.add_argument(
            "--convert-function-body-to-file", help="Dump out FUNCTION body to separate files", default=False, action="store_true"
        )

        parser.add_argument(
            "--convert-view-text-to-file", help="Dump VIEW text to separate files", default=False, action="store_true"
        )

        return parser

    def init_config_path(self):
        config_path = Path(self.args["c"])

        if config_path.is_dir() and self.args.get("clean"):
            rmtree(config_path)

        if not config_path.exists():
            config_path.mkdir(mode=0o755, parents=True)

        if not config_path.is_dir():
            raise ValueError(f"Config path [{self.args['c']}] is not a directory")

        return config_path

    def init_config(self):
        return SnowDDLConfig(self.env_prefix)

    def init_settings(self):
        settings = SnowDDLSettings()

        if self.args.get("exclude_object_types"):
            try:
                settings.exclude_object_types = [
                    ObjectType[t.strip().upper()] for t in str(self.args.get("exclude_object_types")).split(",")
                ]
            except KeyError as e:
                raise ValueError(f"Invalid object type [{str(e)}]")

        if self.args.get("include_object_types"):
            try:
                settings.include_object_types = [
                    ObjectType[t.strip().upper()] for t in str(self.args.get("include_object_types")).split(",")
                ]
            except KeyError as e:
                raise ValueError(f"Invalid object type [{str(e)}]")

        if self.args.get("include_databases"):
            settings.include_databases = [
                DatabaseIdent(self.config.env_prefix, d) for d in str(self.args.get("include_databases")).split(",")
            ]

        if self.args.get("ignore_ownership"):
            settings.ignore_ownership = True

        if self.args.get("add_include_files"):
            settings.add_include_files = True

        if self.args.get("max_workers"):
            settings.max_workers = int(self.args.get("max_workers"))

        if self.args.get("convert_function_body_to_file"):
            settings.convert_function_body_to_file = True

        if self.args.get("convert_view_text_to_file"):
            settings.convert_view_text_to_file = True

        return settings

    def execute(self):
        error_count = 0

        with self.engine:
            self.output_engine_context()

            for converter_cls in default_converter_sequence:
                converter = converter_cls(self.engine, self.config_path)
                converter.convert()

                error_count += len(converter.errors)

            if error_count > 0:
                exit(8)


def entry_point():
    app = ConvertApp()
    app.execute()



from snowddl.blueprint import (
    AccountObjectIdent,
    AuthenticationPolicyBlueprint,
    AuthenticationPolicyReference,
    NetworkPolicyBlueprint,
    NetworkPolicyReference,
    ObjectType,
    SchemaObjectIdent,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
account_policy_json_schema = {
    "type": "object",
    "properties": {
        "authentication_policy": {
            "type": "string"
        },
        "network_policy": {
            "type": "string"
        },
    },
    "additionalProperties": False
}
# fmt: on


class AccountPolicyParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file("account_policy", account_policy_json_schema, self.process_account_policy)

    def process_account_policy(self, f: ParsedFile):
        if self.env_prefix:
            # Account-level policies are ignored with env_prefix is present
            # Can only assign one account-level policy per account, no way around it
            return

        if f.params.get("authentication_policy"):
            policy_name_parts = f.params.get("authentication_policy").split(".")

            if len(policy_name_parts) != 3:
                raise ValueError(
                    f"Authentication policy [{f.params.get('authentication_policy')}] should use fully-qualified identifier <database>.<schema>.<name>"
                )

            policy_name = SchemaObjectIdent(self.env_prefix, *policy_name_parts)

            ref = AuthenticationPolicyReference(
                object_type=ObjectType.ACCOUNT,
            )

            self.config.add_policy_reference(AuthenticationPolicyBlueprint, policy_name, ref)

        if f.params.get("network_policy"):
            policy_name = AccountObjectIdent(self.env_prefix, f.params.get("network_policy"))

            ref = NetworkPolicyReference(
                object_type=ObjectType.ACCOUNT,
            )

            self.config.add_policy_reference(NetworkPolicyBlueprint, policy_name, ref)



from snowddl.blueprint import NetworkRuleBlueprint, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
network_rule_json_schema = {
    "type": "object",
    "properties": {
        "type": {
            "type": "string"
        },
        "value_list": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "mode": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        },
    },
    "required": ["type", "value_list", "mode"],
    "additionalProperties": False
}
# fmt: on


class NetworkRuleParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("network_rule", network_rule_json_schema, self.process_sequence)

    def process_sequence(self, f: ParsedFile):
        bp = NetworkRuleBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            type=str(f.params["type"]).upper(),
            value_list=[str(v) for v in f.params["value_list"]],
            mode=str(f.params["mode"]).upper(),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import ExternalAccessIntegrationBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class ExternalAccessIntegrationResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.EXTERNAL_ACCESS_INTEGRATION

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW EXTERNAL ACCESS INTEGRATIONS LIKE {env_prefix:ls}",
            {
                "env_prefix": self.config.env_prefix,
            },
        )

        for r in cur:
            existing_objects[r["name"]] = {
                "name": r["name"],
                "enabled": r["enabled"],
                "comment": r["comment"] if r["comment"] else None,
            }

        for name, owner in self.engine.executor.map(self.get_owner_from_grant, existing_objects.keys()):
            if owner != self.engine.context.current_role:
                del existing_objects[name]

        return existing_objects

    def get_owner_from_grant(self, name):
        cur = self.engine.execute_meta(
            "SHOW GRANTS ON INTEGRATION {name:i}",
            {
                "name": name,
            },
        )

        for r in cur:
            # Assumption: OWNERSHIP grant always exist, exactly 1 row
            if r["privilege"] == "OWNERSHIP":
                return name, r["grantee_name"]

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ExternalAccessIntegrationBlueprint)

    def create_object(self, bp: ExternalAccessIntegrationBlueprint):
        common_query = self._build_common_external_access_integration_sql(bp)
        create_query = self.engine.query_builder()

        create_query.append(
            "CREATE EXTERNAL ACCESS INTEGRATION {name:i}",
            {
                "name": bp.full_name,
            },
        )

        create_query.append_nl(common_query)
        self.engine.execute_safe_ddl(create_query)

        self.engine.execute_safe_ddl(
            "COMMENT ON INTEGRATION {name:i} IS {comment}",
            {
                "name": bp.full_name,
                "comment": common_query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: ExternalAccessIntegrationBlueprint, row: dict):
        common_query = self._build_common_external_access_integration_sql(bp)

        if not common_query.compare_short_hash(row["comment"]):
            replace_query = self.engine.query_builder()

            replace_query.append(
                "CREATE OR REPLACE EXTERNAL ACCESS INTEGRATION {name:i}",
                {
                    "name": bp.full_name,
                },
            )

            replace_query.append_nl(common_query)
            self.engine.execute_safe_ddl(replace_query)

            self.engine.execute_safe_ddl(
                "COMMENT ON INTEGRATION {name:i} IS {comment}",
                {
                    "name": bp.full_name,
                    "comment": common_query.add_short_hash(bp.comment),
                },
            )

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP EXTERNAL ACCESS INTEGRATION {name:i}",
            {
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_external_access_integration_sql(self, bp: ExternalAccessIntegrationBlueprint):
        query = self.engine.query_builder()

        query.append_nl(
            "ALLOWED_NETWORK_RULES = ({allowed_network_rules:i})",
            {
                "allowed_network_rules": bp.allowed_network_rules,
            },
        )

        if bp.allowed_api_authentication_integrations:
            query.append_nl(
                "ALLOWED_API_AUTHENTICATION_INTEGRATIONS = ({allowed_api_authentication_integrations:i})",
                {
                    "allowed_api_authentication_integrations": bp.allowed_api_authentication_integrations,
                },
            )

        if bp.allowed_authentication_secrets:
            query.append_nl(
                "ALLOWED_AUTHENTICATION_SECRETS = ({allowed_authentication_secrets:i})",
                {
                    "allowed_authentication_secrets": bp.allowed_authentication_secrets,
                },
            )

        query.append_nl(
            "ENABLED = {enabled:b}",
            {
                "enabled": bp.enabled,
            },
        )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query



from snowddl.blueprint import FunctionBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import dtypes_from_arguments


class FunctionResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.FUNCTION

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW USER FUNCTIONS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            if r["is_external_function"] == "Y" or r["is_data_metric"] == "Y":
                continue

            full_name = f"{r['catalog_name']}.{r['schema_name']}.{r['name']}({dtypes_from_arguments(r['arguments'])})"

            existing_objects[full_name] = {
                "database": r["catalog_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "arguments": r["arguments"],
                "comment": r["description"],
                "is_table_function": r["is_table_function"] == "Y",
                "is_secure": r["is_secure"] == "Y",
                "is_memoizable": r["is_memoizable"] == "Y",
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(FunctionBlueprint)

    def create_object(self, bp: FunctionBlueprint):
        query = self._build_create_function(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl(
            "COMMENT ON FUNCTION {full_name:i} IS {comment}",
            {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: FunctionBlueprint, row: dict):
        query = self._build_create_function(bp)

        if not query.compare_short_hash(row["comment"]):
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl(
                "COMMENT ON FUNCTION {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": query.add_short_hash(bp.comment),
                },
            )

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "DROP FUNCTION {database:i}.{schema:i}.{name:i}({dtypes:r})",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
                "dtypes": dtypes_from_arguments(row["arguments"]),
            },
        )

        return ResolveResult.DROP

    def _build_create_function(self, bp: FunctionBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE OR REPLACE")

        if bp.is_secure:
            query.append("SECURE")

        query.append(
            "FUNCTION {full_name:in} (",
            {
                "full_name": bp.full_name,
            },
        )

        for idx, arg in enumerate(bp.arguments):
            query.append_nl(
                "    {comma:r}{arg_name:i} {arg_type:r}",
                {
                    "comma": "  " if idx == 0 else ", ",
                    "arg_name": arg.name,
                    "arg_type": arg.type,
                },
            )

            if arg.default:
                query.append(
                    "DEFAULT {default:r}",
                    {
                        "default": arg.default,
                    },
                )

        query.append_nl(")")

        if isinstance(bp.returns, list):
            query.append_nl("RETURNS TABLE (")

            for idx, arg in enumerate(bp.returns):
                query.append_nl(
                    "    {comma:r}{ret_name:i} {ret_type:r}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "ret_name": arg.name,
                        "ret_type": arg.type,
                    },
                )

            query.append_nl(")")
        else:
            query.append_nl(
                "RETURNS {ret_type:r}",
                {
                    "ret_type": bp.returns,
                },
            )

        query.append_nl(
            "LANGUAGE {language:r}",
            {
                "language": bp.language,
            },
        )

        if bp.is_strict:
            query.append_nl("STRICT")

        if bp.is_immutable:
            query.append_nl("IMMUTABLE")

        if bp.is_memoizable:
            query.append_nl("MEMOIZABLE")

        if bp.runtime_version:
            query.append_nl(
                "RUNTIME_VERSION = {runtime_version}",
                {
                    "runtime_version": bp.runtime_version,
                },
            )

        if bp.imports:
            query.append_nl("IMPORTS = ({imports})", {"imports": [f"@{i.stage_name}{i.path}" for i in bp.imports]})

        if bp.packages:
            query.append_nl(
                "PACKAGES = ({packages})",
                {
                    "packages": bp.packages,
                },
            )

        if bp.handler:
            query.append_nl(
                "HANDLER = {handler}",
                {
                    "handler": bp.handler,
                },
            )

        if bp.external_access_integrations:
            # Snowflake bug: EXTERNAL_ACCESS_INTEGRATIONS must be identifiers
            # It does not accept identifiers in double-quotes
            query.append_nl(
                "EXTERNAL_ACCESS_INTEGRATIONS = ({external_access_integrations:r})",
                {
                    "external_access_integrations": bp.external_access_integrations,
                },
            )

        if bp.secrets:
            query.append_nl("SECRETS = (")

            for idx, (var_name, secret_name) in enumerate(bp.secrets.items()):
                query.append(
                    "{comma:r}{var_name} = {secret_name:i}",
                    {
                        "comma": "" if idx == 0 else ", ",
                        "var_name": var_name,
                        "secret_name": secret_name,
                    },
                )

            query.append(")")

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        if bp.body:
            query.append_nl(
                "AS {body}",
                {
                    "body": bp.body,
                },
            )

        return query



from argparse import ArgumentParser, HelpFormatter
from contextlib import contextmanager
from cryptography.hazmat.primitives import serialization
from importlib.util import module_from_spec, spec_from_file_location
from json import loads as json_loads
from json.decoder import JSONDecodeError
from logging import getLogger, Formatter, StreamHandler
from os import environ, getcwd
from pathlib import Path
from snowflake.connector import connect
from string import ascii_uppercase, digits
from time import perf_counter
from traceback import TracebackException

from snowddl.blueprint import Ident, ObjectType
from snowddl.config import SnowDDLConfig
from snowddl.engine import SnowDDLEngine
from snowddl.parser import default_parse_sequence, DirectoryScanner, PermissionModelParser, PlaceholderParser
from snowddl.resolver import default_resolve_sequence, default_destroy_sequence
from snowddl.settings import SnowDDLSettings
from snowddl.version import __version__


class BaseApp:
    application_name = "SnowDDL"
    application_version = __version__

    parse_sequence = default_parse_sequence
    resolve_sequence = default_resolve_sequence
    destroy_sequence = default_destroy_sequence

    def __init__(self):
        self.elapsed_timers = {}

        self.arg_parser = self.init_arguments_parser()
        self.args = self.init_arguments()
        self.logger = self.init_logger()

        with self.measure_elapsed_time("InitConfig"):
            self.env_prefix = self.init_env_prefix()
            self.config_path = self.init_config_path()
            self.config = self.init_config()
            self.settings = self.init_settings()

        with self.measure_elapsed_time("InitEngine"):
            self.engine = self.init_engine()

    def init_arguments_parser(self):
        formatter = lambda prog: HelpFormatter(prog, max_help_position=36)

        parser = ArgumentParser(
            prog="snowddl", description="Object management automation tool for Snowflake", formatter_class=formatter
        )

        # Config
        parser.add_argument(
            "-c",
            help="Path to config directory OR name of bundled test config (default: current directory)",
            metavar="CONFIG_PATH",
            default=getcwd(),
        )

        # Auth
        parser.add_argument(
            "-a",
            help="Snowflake account identifier (default: SNOWFLAKE_ACCOUNT env variable)",
            metavar="ACCOUNT",
            default=environ.get("SNOWFLAKE_ACCOUNT"),
        )
        parser.add_argument(
            "-u",
            help="Snowflake user name (default: SNOWFLAKE_USER env variable)",
            metavar="USER",
            default=environ.get("SNOWFLAKE_USER"),
        )
        parser.add_argument(
            "-p",
            help="Snowflake user password (default: SNOWFLAKE_PASSWORD env variable)",
            metavar="PASSWORD",
            default=environ.get("SNOWFLAKE_PASSWORD"),
        )
        parser.add_argument(
            "-k",
            help="Path to private key file (default: SNOWFLAKE_PRIVATE_KEY_PATH env variable)",
            metavar="PRIVATE_KEY",
            default=environ.get("SNOWFLAKE_PRIVATE_KEY_PATH"),
        )

        # Role & warehouse
        parser.add_argument(
            "-r",
            help="Snowflake active role (default: SNOWFLAKE_ROLE env variable)",
            metavar="ROLE",
            default=environ.get("SNOWFLAKE_ROLE"),
        )
        parser.add_argument(
            "-w",
            help="Snowflake active warehouse (default: SNOWFLAKE_WAREHOUSE env variable)",
            metavar="WAREHOUSE",
            default=environ.get("SNOWFLAKE_WAREHOUSE"),
        )

        # Options
        parser.add_argument(
            "--authenticator",
            help="Authenticator: 'snowflake' or 'externalbrowser' (to use any IdP and a web browser) (default: SNOWFLAKE_AUTHENTICATOR env variable or 'snowflake')",
            default=environ.get("SNOWFLAKE_AUTHENTICATOR", "snowflake"),
        )
        parser.add_argument(
            "--passphrase",
            help="Passphrase for private key file (default: SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable)",
            default=environ.get("SNOWFLAKE_PRIVATE_KEY_PASSPHRASE"),
        )
        parser.add_argument(
            "--env-prefix",
            help="Env prefix added to global object names, used to separate environments (e.g. DEV, PROD)",
            default=environ.get("SNOWFLAKE_ENV_PREFIX"),
        )
        parser.add_argument(
            "--env-prefix-separator",
            help="Custom separator for Env prefix (supported values are: '__', '_', '$')",
            choices=["__", "_", "$"],
            default=environ.get("SNOWFLAKE_ENV_PREFIX_SEPARATOR", "__"),
        )
        parser.add_argument(
            "--env-admin-role",
            help="Super administration role which should inherit env prefixed SnowDDL role",
            default=environ.get("SNOWFLAKE_ENV_ADMIN_ROLE"),
        )
        parser.add_argument(
            "--max-workers", help="Maximum number of workers to resolve objects in parallel", default=None, type=int
        )
        parser.add_argument(
            "--query-tag",
            help="Add QUERY_TAG to all queries produced by SnowDDL",
            default=environ.get("SNOWFLAKE_QUERY_TAG"),
        )

        # Logging
        parser.add_argument(
            "--log-level", help="Log level (possible values: DEBUG, INFO, WARNING; default: INFO)", default="INFO"
        )
        # fmt: off
        parser.add_argument(
            "--show-sql", help="Show executed DDL queries", default=False, action="store_true"
        )
        parser.add_argument(
            "--show-timers", help="Show debug timers", default=False, action="store_true"
        )
        # fmt: on
        parser.add_argument(
            "--show-unused-files", help="Show warnings for unused config files", default=False, action="store_true"
        )

        # Placeholders
        parser.add_argument(
            "--placeholder-path", help="Path to config file with environment-specific placeholders", default=None, metavar=""
        )
        parser.add_argument(
            "--placeholder-values", help="Environment-specific placeholder values in JSON format", default=None, metavar=""
        )

        # Object types
        parser.add_argument(
            "--exclude-object-types", help="Comma-separated list of object types NOT to resolve", default=None, metavar=""
        )
        parser.add_argument(
            "--include-object-types",
            help="Comma-separated list of object types TO resolve, all other types are excluded",
            default=None,
            metavar="",
        )

        # Apply even more unsafe changes
        parser.add_argument(
            "--apply-unsafe",
            help="Additionally apply unsafe changes, which may cause loss of data (ALTER, DROP, etc.)",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-replace-table",
            help="Additionally apply REPLACE TABLE when ALTER TABLE is not possible",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-all-policy", help="Additionally apply changes for all types of policies", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-account-level-policy",
            help="Additionally apply changes for ACCOUNT-level policies",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-aggregation-policy",
            help="Additionally apply changes to AGGREGATION POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-authentication-policy",
            help="Additionally apply changes to AUTHENTICATION POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-masking-policy", help="Additionally apply changes to MASKING POLICIES", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-projection-policy",
            help="Additionally apply changes to PROJECTION POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-row-access-policy",
            help="Additionally apply changes to ROW ACCESS POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-account-params", help="Additionally apply changes to ACCOUNT PARAMETERS", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-network-policy", help="Additionally apply changes to NETWORK POLICIES", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-resource-monitor", help="Additionally apply changes to RESOURCE MONITORS", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-outbound-share", help="Additionally apply changes to OUTBOUND SHARES", default=False, action="store_true"
        )

        # Refresh state of specific objects
        parser.add_argument(
            "--refresh-user-passwords", help="Additionally refresh passwords of users", default=False, action="store_true"
        )
        parser.add_argument(
            "--refresh-future-grants",
            help="Additionally refresh missing grants for existing objects derived from future grants",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--refresh-stage-encryption",
            help="Additionally refresh stage encryption parameters for existing external stages",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--refresh-secrets",
            help="Additionally refresh secrets",
            default=False,
            action="store_true",
        )

        # Cloning
        parser.add_argument(
            "--clone-table",
            help="Clone all tables from source databases (without env_prefix) to destination databases (with env_prefix)",
            default=False,
            action="store_true",
        )

        # Destroy without env prefix
        parser.add_argument(
            "--destroy-without-prefix", help="Allow {destroy} action without --env-prefix", default=False, action="store_true"
        )

        # Subparsers
        subparsers = parser.add_subparsers(dest="action")
        subparsers.required = True

        subparsers.add_parser("plan", help="Resolve objects, apply nothing, display suggested changes")
        subparsers.add_parser("apply", help="Resolve objects, apply safe changes, display suggested unsafe changes")
        subparsers.add_parser(
            "destroy", help="Drop objects with specified --env-prefix, use it to reset dev and test environments"
        )

        return parser

    def init_arguments(self):
        args = vars(self.arg_parser.parse_args())

        if not self.validate_auth_args(args):
            self.arg_parser.print_help()
            exit(1)

        return args

    def validate_auth_args(self, args):
        if args["authenticator"] == "snowflake":
            if not args["a"] or not args["u"] or (not args["p"] and not args["k"] and "SNOWFLAKE_PRIVATE_KEY" not in environ):
                return False
        elif args["authenticator"] == "externalbrowser":
            if not args["a"] or not args["u"]:
                return False
        elif args["authenticator"] is not None:
            return False
        return True

    def init_logger(self):
        logger = getLogger("snowddl")
        logger.setLevel(self.args.get("log_level", "INFO"))

        formatter = Formatter("%(asctime)s - %(levelname)s - %(message)s")
        formatter.default_msec_format = "%s.%03d"

        handler = StreamHandler()
        handler.setFormatter(formatter)

        logger.addHandler(handler)

        return logger

    def init_env_prefix(self):
        env_prefix_value = self.args.get("env_prefix")
        env_prefix_separator = self.args.get("env_prefix_separator")

        allowed_env_prefix_value_chars = set(ascii_uppercase + digits + "_")

        if env_prefix_value:
            env_prefix_value = str(env_prefix_value).upper()

            for char in env_prefix_value:
                if char not in allowed_env_prefix_value_chars:
                    raise ValueError(
                        f"Character [{char}] in not allowed in env prefix [{env_prefix_value}], only ASCII letters, digits and single underscores are accepted"
                    )

            if env_prefix_separator in env_prefix_value:
                raise ValueError(f"Env prefix [{env_prefix_value}] cannot contain env prefix separator [{env_prefix_separator}]")

            if env_prefix_value.endswith("_"):
                raise ValueError(f"Env prefix [{env_prefix_value}] cannot end with [_] underscore")

            return f"{env_prefix_value}{env_prefix_separator}"

        return ""

    def init_config_path(self):
        config_path = Path(self.args["c"])

        if not config_path.exists():
            config_path = Path(__file__).parent.parent / "_config" / self.args["c"]

        if not config_path.exists():
            raise ValueError(f"Config path [{self.args['c']}] does not exist")

        if not config_path.is_dir():
            raise ValueError(f"Config path [{self.args['c']}] is not a directory")

        return config_path.resolve()

    def init_config(self):
        config = SnowDDLConfig(self.env_prefix)
        scanner = DirectoryScanner(self.config_path)

        # Placeholders
        placeholder_path = self.get_placeholder_path()
        placeholder_values = self.get_placeholder_values()

        parser = PlaceholderParser(config, scanner)
        parser.load_placeholders(placeholder_path, placeholder_values, self.args)

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        # Permission models
        parser = PermissionModelParser(config, scanner)
        parser.load_permission_models()

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        # Blueprints
        for parser_cls in self.parse_sequence:
            parser = parser_cls(config, scanner)
            parser.load_blueprints()

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        # Custom programmatically generated blueprints and config adjustments
        for module_path in sorted(self.config_path.glob("__custom/*.py")):
            try:
                spec = spec_from_file_location(module_path.name, module_path)

                module = module_from_spec(spec)
                spec.loader.exec_module(module)

                module.handler(config)
            except Exception as e:
                config.add_error(module_path, e)

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        if self.args.get("show_unused_files"):
            self.output_unused_file_warnings(scanner.get_unused_file_paths())

        return config

    def init_settings(self):
        settings = SnowDDLSettings()

        if self.args.get("action") in ("apply", "destroy"):
            settings.execute_safe_ddl = True

            if self.args.get("apply_unsafe") or self.args.get("action") == "destroy":
                settings.execute_unsafe_ddl = True

            if self.args.get("apply_replace_table"):
                settings.execute_replace_table = True

            if self.args.get("apply_all_policy"):
                settings.execute_account_level_policy = True
                settings.execute_aggregation_policy = True
                settings.execute_authentication_policy = True
                settings.execute_masking_policy = True
                settings.execute_projection_policy = True
                settings.execute_row_access_policy = True
                settings.execute_network_policy = True

            if self.args.get("apply_account_level_policy"):
                settings.execute_account_level_policy = True

            if self.args.get("apply_aggregation_policy"):
                settings.execute_aggregation_policy = True

            if self.args.get("apply_authentication_policy"):
                settings.execute_authentication_policy = True

            if self.args.get("apply_masking_policy"):
                settings.execute_masking_policy = True

            if self.args.get("apply_projection_policy"):
                settings.execute_projection_policy = True

            if self.args.get("apply_row_access_policy"):
                settings.execute_row_access_policy = True

            if self.args.get("apply_account_params"):
                settings.execute_account_params = True

            if self.args.get("apply_network_policy"):
                settings.execute_network_policy = True

            if self.args.get("apply_resource_monitor"):
                settings.execute_resource_monitor = True

            if self.args.get("apply_outbound_share"):
                settings.execute_outbound_share = True

        if self.args.get("refresh_user_passwords"):
            settings.refresh_user_passwords = True

        if self.args.get("refresh_future_grants"):
            settings.refresh_future_grants = True

        if self.args.get("refresh_stage_encryption"):
            settings.refresh_stage_encryption = True

        if self.args.get("refresh_secrets"):
            settings.refresh_secrets = True

        if self.args.get("clone_table"):
            if self.args.get("action") != "apply":
                raise ValueError("Argument --clone-table requires action [apply]")

            if not self.args.get("env_prefix"):
                raise ValueError("Argument --clone-table requires argument --env-prefix")

            settings.clone_table = True

        if self.args.get("env_admin_role"):
            settings.env_admin_role = Ident(self.args.get("env_admin_role"))

        if self.args.get("exclude_object_types"):
            try:
                settings.exclude_object_types = [
                    ObjectType[t.strip().upper()] for t in str(self.args.get("exclude_object_types")).split(",")
                ]
            except KeyError as e:
                raise ValueError(f"Invalid object type [{str(e)}]")

        if self.args.get("include_object_types"):
            try:
                settings.include_object_types = [
                    ObjectType[t.strip().upper()] for t in str(self.args.get("include_object_types")).split(",")
                ]
            except KeyError as e:
                raise ValueError(f"Invalid object type [{str(e)}]")

        if self.args.get("max_workers"):
            settings.max_workers = int(self.args.get("max_workers"))

        return settings

    def init_engine(self):
        return SnowDDLEngine(self.get_connection(), self.config, self.settings)

    def get_connection(self):
        options = {
            "account": self.args["a"],
            "user": self.args["u"],
            "role": self.args["r"],
            "warehouse": self.args["w"],
            "application": f"{self.application_name} {self.application_version}",
        }

        if self.args.get("authenticator") == "snowflake":
            key_bytes = None

            if self.args.get("k"):
                key_path = Path(self.args.get("k"))

                if not key_path.is_file():
                    raise ValueError(f"Private key file [{key_path}] does not exist or not a file")

                key_bytes = key_path.read_bytes()
            elif "SNOWFLAKE_PRIVATE_KEY" in environ:
                key_bytes = str(environ["SNOWFLAKE_PRIVATE_KEY"]).encode("utf-8")

            if key_bytes:
                key_password = str(self.args.get("passphrase")).encode("utf-8") if self.args.get("passphrase") else None
                pk = serialization.load_pem_private_key(data=key_bytes, password=key_password)

                options["private_key"] = pk.private_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            else:
                options["password"] = self.args["p"]
        elif self.args.get("authenticator") == "externalbrowser":
            options["authenticator"] = "externalbrowser"
        else:
            raise ValueError("Only 'snowflake' and 'externalbrowser' authenticators are supported")

        if self.args.get("query_tag"):
            options["session_parameters"] = {
                "QUERY_TAG": self.args.get("query_tag"),
            }

        return connect(**options)

    def execute(self):
        error_count = 0

        with self.engine:
            self.output_engine_context()

            if self.args.get("action") == "destroy":
                if not self.args.get("env_prefix") and not self.args.get("destroy_without_prefix"):
                    raise ValueError("Argument --env-prefix is required for [destroy] action")

                for resolver_cls in self.destroy_sequence:
                    with self.measure_elapsed_time(resolver_cls.__name__):
                        resolver = resolver_cls(self.engine)
                        resolver.destroy()

                    error_count += len(resolver.errors)

                self.engine.context.destroy_role_with_prefix()
            else:
                for resolver_cls in self.resolve_sequence:
                    with self.measure_elapsed_time(resolver_cls.__name__):
                        resolver = resolver_cls(self.engine)
                        resolver.resolve()

                    error_count += len(resolver.errors)

            self.engine.connection.close()
            self.output_engine_stats()
            self.output_engine_warnings()

            if self.args.get("show_timers"):
                self.output_app_timers()

            if self.args.get("show_sql"):
                self.output_executed_ddl()

            self.output_suggested_ddl()

            if error_count > 0:
                exit(8)

    def output_engine_context(self):
        roles = []

        if self.engine.context.is_account_admin:
            roles.append("ACCOUNTADMIN")

        if self.engine.context.is_sys_admin:
            roles.append("SYSADMIN")

        if self.engine.context.is_security_admin:
            roles.append("SECURITYADMIN")

        self.logger.info(
            f"Snowflake version = {self.engine.context.version} ({self.engine.context.edition.name}), SnowDDL version = {__version__}"
        )
        self.logger.info(f"Account = {self.engine.context.current_account}, Region = {self.engine.context.current_region}")
        self.logger.info(f"Session = {self.engine.context.current_session}, User = {self.engine.context.current_user}")
        self.logger.info(f"Role = {self.engine.context.current_role}, Warehouse = {self.engine.context.current_warehouse}")
        self.logger.info(f"Roles in session = {','.join(roles)}")
        self.logger.info("---")

    def get_placeholder_path(self):
        if self.args.get("placeholder_path"):
            placeholder_path = Path(self.args.get("placeholder_path"))

            if not placeholder_path.is_file():
                raise ValueError(f"Placeholder path [{self.args.get('placeholder_path')}] does not exist or not a file")

            return placeholder_path.resolve()

        return None

    def get_placeholder_values(self):
        if self.args.get("placeholder_values"):
            try:
                placeholder_values = json_loads(self.args.get("placeholder_values"))
            except JSONDecodeError:
                raise ValueError(f"Placeholder values [{self.args.get('placeholder_values')}] are not a valid JSON")

            if not isinstance(placeholder_values, dict):
                raise ValueError(f"Placeholder values [{self.args.get('placeholder_values')}] are not JSON encoded dict")

            for k, v in placeholder_values.items():
                if isinstance(v, list):
                    for item in v:
                        if not isinstance(item, (bool, float, int, str)):
                            raise ValueError(
                                f"Invalid type [{type(item).__name__}] of placeholder [{k.upper()}] item, supported types are: bool, float, int, str"
                            )

                elif not isinstance(v, (bool, float, int, str)):
                    raise ValueError(
                        f"Invalid type [{type(v).__name__}] of placeholder [{k.upper()}], supported types are scalars or lists of: bool, float, int, str"
                    )

            return placeholder_values

        return None

    def output_config_errors(self, config):
        for e in config.errors:
            self.logger.warning(f"[{e['path']}]: {''.join(TracebackException.from_exception(e['error']).format())}")

    def output_engine_stats(self):
        self.logger.info(
            f"Executed {len(self.engine.executed_ddl)} DDL queries, Suggested {len(self.engine.suggested_ddl)} DDL queries"
        )

    def output_engine_warnings(self):
        for object_type, object_names in self.engine.intention_cache.invalid_name_warning.items():
            for name in object_names:
                self.logger.warning(
                    f"Detected {object_type.name} with name [{name}] "
                    f"which does not conform to SnowDDL standards, please rename or drop it manually"
                )

    def output_unused_file_warnings(self, unused_file_paths):
        for file_path in unused_file_paths.values():
            self.logger.warning(f"Detected possibly unused config file [{file_path}]")

    def output_app_timers(self):
        for timer_name, timer_value in self.elapsed_timers.items():
            self.logger.info(f"Timer [{timer_name}] elapsed time is {timer_value:.3f}s")

    def output_suggested_ddl(self):
        if self.engine.suggested_ddl:
            print("--- Suggested DDL ---\n")

        for sql in self.engine.suggested_ddl:
            print(f"{sql};\n")

    def output_executed_ddl(self):
        if self.engine.executed_ddl:
            print("--- Executed DDL ---\n")

        for sql in self.engine.executed_ddl:
            print(f"{sql};\n")

    @contextmanager
    def measure_elapsed_time(self, timer_name: str):
        start_counter = perf_counter()

        try:
            yield
        finally:
            self.elapsed_timers[timer_name] = perf_counter() - start_counter


def entry_point():
    app = BaseApp()
    app.execute()



from abc import abstractmethod

from snowddl.blueprint import SchemaBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class AbstractSchemaObjectResolver(AbstractResolver):
    def get_existing_objects(self):
        existing_objects = {}

        # Process schemas in parallel
        for schema_objects in self.engine.executor.map(
            self.get_existing_objects_in_schema, self.engine.schema_cache.schemas.values()
        ):
            existing_objects.update(schema_objects)

        return existing_objects

    @abstractmethod
    def get_existing_objects_in_schema(self, schema: dict):
        pass

    def _resolve_drop(self):
        tasks = {}

        for object_full_name in sorted(self.existing_objects):
            # Object exists in blueprints, should not be dropped
            if object_full_name in self.blueprints:
                continue

            # Parent object is going to be dropped
            if self.engine.intention_cache.check_parent_drop_intention(self.object_type, object_full_name):
                continue

            schema_full_name = ".".join(object_full_name.split(".")[:2])
            schema_bp = self.config.get_blueprints_by_type(SchemaBlueprint).get(schema_full_name)

            # Object schema does not exist in blueprints, object will be dropped automatically on DROP DATABASE or DROP SCHEMA
            if schema_bp is None:
                continue

            # Objects without blueprints are allowed in sandbox schemas, should not be dropped
            if schema_bp.is_sandbox:
                continue

            tasks[object_full_name] = (self.drop_object, self.existing_objects[object_full_name])

        self._process_tasks(tasks)



from cryptography.fernet import Fernet, MultiFernet
from os import environ
from typing import List, Optional


class FernetWrapper:
    ENV_ENCRYPTION_KEYS = "SNOWFLAKE_CONFIG_FERNET_KEYS"

    def __init__(self, explicit_encryption_keys: Optional[str] = None):
        self.key_sequence: List[str] = []
        self.multi_fernet: Optional[MultiFernet] = None

        if explicit_encryption_keys:
            self.key_sequence = self._parse_encryption_keys(explicit_encryption_keys)
        elif self.ENV_ENCRYPTION_KEYS in environ:
            self.key_sequence = self._parse_encryption_keys(environ[self.ENV_ENCRYPTION_KEYS])

        if self.key_sequence:
            self.multi_fernet = MultiFernet(Fernet(key) for key in self.key_sequence)

    def generate_key(self):
        return Fernet.generate_key().decode("ascii")

    def encrypt(self, value: str):
        if not self.multi_fernet:
            raise RuntimeError("Cannot encrypt value due to missing Fernet encryption keys")

        return self.multi_fernet.encrypt(value.encode("utf-8")).decode("ascii")

    def decrypt(self, value: str):
        if not self.multi_fernet:
            raise RuntimeError("Cannot decrypt value due to missing Fernet encryption keys")

        return self.multi_fernet.decrypt(value.encode("ascii")).decode("utf-8")

    def rotate(self, value: str):
        if not self.multi_fernet:
            raise RuntimeError("Cannot rotate value due to missing Fernet encryption keys")

        return self.multi_fernet.rotate(value.encode("ascii")).decode("ascii")

    def _parse_encryption_keys(self, encryption_keys: str):
        key_sequence = []

        for key in encryption_keys.split(","):
            key = key.strip()

            if not key:
                continue

            key_sequence.append(key)

        return key_sequence



from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Set


class DirectoryScanner:
    allowed_file_suffixes = (".yml", ".yaml")
    allowed_file_levels = (1, 2, 3, 4)

    def __init__(self, base_path: Path):
        self.base_path = base_path

        # Key is relative path without suffix
        self.all_file_paths: Dict[str, Path] = {}
        self.accessed_files: Set[str] = set()

        # Key is database name
        self.database_dir_paths: Dict[str, Path] = {}

        # Key 1st level -> database name, 2nd level -> schema name
        self.schema_dir_paths: Dict[str, Dict[str, Path]] = defaultdict(dict)

        # Key 1st level -> object type, 2nd level -> relative path without suffix
        self.schema_object_file_paths: Dict[str, Dict[str, Path]] = defaultdict(dict)

        self._scan_dir_recursively(base_path)

    def get_database_dir_paths(self) -> Dict[str, Path]:
        return self.database_dir_paths

    def get_schema_dir_paths(self, database_key) -> Dict[str, Path]:
        return self.schema_dir_paths[self._normalise_key(database_key)]

    def get_single_file_path(self, file_key: str) -> Optional[Path]:
        file_key = self._normalise_key(file_key)

        if file_key in self.all_file_paths:
            self.accessed_files.add(file_key)
            return self.all_file_paths[file_key]

        return None

    def get_schema_object_file_paths(self, object_type: str) -> Dict[str, Path]:
        schema_object_file_paths = self.schema_object_file_paths.get(object_type, {})

        for file_key in schema_object_file_paths:
            self.accessed_files.add(file_key)

        return schema_object_file_paths

    def get_unused_file_paths(self) -> Dict[str, Path]:
        return {file_key: file_path for file_key, file_path in self.all_file_paths.items() if file_key not in self.accessed_files}

    def _scan_dir_recursively(self, dir_path: Path):
        for item_path in sorted(dir_path.iterdir()):
            relative_item_path = item_path.relative_to(self.base_path)
            relative_item_level = len(relative_item_path.parts)

            if relative_item_path.parts[0].startswith("__"):
                # Skip special directories and all items inside
                continue

            if item_path.is_dir():
                # 1st level dir is database
                if relative_item_level == 1:
                    database_key = self._normalise_key(relative_item_path.parts[0])

                    if database_key in self.database_dir_paths:
                        raise ValueError(
                            f"Detected config database directories with similar names: "
                            f"[{item_path}] and [{self.database_dir_paths[database_key]}]"
                        )

                    self.database_dir_paths[database_key] = item_path

                # 2nd level dir is schema
                if relative_item_level == 2:
                    database_key = self._normalise_key(relative_item_path.parts[0])
                    schema_key = self._normalise_key(relative_item_path.parts[1])

                    if database_key in self.schema_dir_paths and schema_key in self.schema_dir_paths[database_key]:
                        raise ValueError(
                            f"Detected config schema directories with similar names: "
                            f"[{item_path}] and [{self.schema_dir_paths[database_key][schema_key]}]"
                        )

                    self.schema_dir_paths[database_key][schema_key] = item_path

                self._scan_dir_recursively(item_path)

            if item_path.is_file():
                # Skip files with extensions other than YML / YAML
                if item_path.suffix not in self.allowed_file_suffixes:
                    continue

                # Skip files outside levels between 1st and 4th
                if relative_item_level not in self.allowed_file_levels:
                    continue

                file_key = self._normalise_key(relative_item_path.with_suffix("").as_posix())

                if file_key in self.all_file_paths:
                    raise ValueError(
                        f"Detected config files with similar names: " f"[{item_path}] and [{self.all_file_paths[file_key]}]"
                    )

                if relative_item_level == 4:
                    object_type = self._normalise_key(relative_item_path.parts[2])
                    self.schema_object_file_paths[object_type][file_key] = item_path

                self.all_file_paths[file_key] = item_path

    def _normalise_key(self, key: str):
        return key.lower()
